/*   This file is part of IndustryConstructor, which is a GameScript for OpenTTD
 *   Copyright (C) 2013  R2dical
 *
 *   This program is free software: you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation, either version 3 of the License, or
 *   (at your option) any later version.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

// Contributors:

// ToDo:

// Error handlers
// Clean code comments
// Debugging support


// Notes:
// Log levels:
//	static LVL_INFO = 1           // main info. eg what it is doing
//	static LVL_SUB_DECISIONS = 2  // sub decisions - eg. reasons for not doing certain things etc.
//	static LVL_DEBUG = 3          // debug prints - debug prints during carrying out actions

// Imports
import("util.superlib", "SuperLib", 36)
Result 		<- SuperLib.Result
Log 		<- SuperLib.Log
Helper 		<- SuperLib.Helper
ScoreList 	<- SuperLib.ScoreList
Tile 		<- SuperLib.Tile
Direction 	<- SuperLib.Direction
Town 		<- SuperLib.Town
Industry	<- SuperLib.Industry

require("common.nut")

///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

// This auto - categorising was intended to be replacement to duplications in script, 
// partially have good results, but not fully finished
const SMITHY_FORGE_NAME = "Smithy Forge"
const BANK_NAME = "Bank"
const WATER_TOWER_NAME = "Water Tower"
const BREWERY_NAME = "Brewery"
const GROCERY_STORE_NAME = "Grocery Store"
const HARDWARE_STORE_NAME = "Hardware Store"
const BUILDERS_YARD_NAME = "Builders Yard"
const HOTEL_NAME = "Hotel"
const FLOUR_MILL_NAME = "Flour Mill"

// Generator mostly is failing, can be fixed, but prospecting almost same
CITY_INDUSTRIES <- [SMITHY_FORGE_NAME, BANK_NAME, WATER_TOWER_NAME, HARDWARE_STORE_NAME, BUILDERS_YARD_NAME, HOTEL_NAME, GROCERY_STORE_NAME]
NEAR_CITY_INDUSTRIES <- [FLOUR_MILL_NAME, BREWERY_NAME]


const OIL_RIG_NAME = "Oil Rig"
const FISHING_GROUNDS_NAME = "Fishing Grounds"
const DREDGING_SITE_NAME = "Dredging Site"

SEA_INDUSTRIES <- [OIL_RIG_NAME, FISHING_GROUNDS_NAME, DREDGING_SITE_NAME]

const FISHING_HARBOUR_NAME = "Fishing Harbor"
const BULK_TERMINAL_NAME = "Bulk Terminal"
const PORT_NAME = "Port"
const BEACH_NAME = "Beach"

// Generator mostly is failing, can be fixed, but prospecting mostly good lazy alternative
COAST_INDUSTRIES <- [FISHING_HARBOUR_NAME, BULK_TERMINAL_NAME, PORT_NAME, BEACH_NAME]

const LUMBER_MILL_NAME = "Lumber Mill"
const CLAY_PIT_NAME = "Clay Pit"
const OIL_REF_NAME = "Oil Refinery"
const FARM_NAME = "Farm"
const FOREST_NAME = "Forest"
const WATER_SUPPLY_NAME = "Water Supply"


DESERT_ONLY_INDUSRIES <- [WATER_SUPPLY_NAME]
TEMP_ONLY_INDUSTRIES <- [CLAY_PIT_NAME, FARM_NAME, FOREST_NAME]

// If fails for unclear reason, this "temporary" override list may help
PROSPECT_ONLY_INDUSTRIES <- CITY_INDUSTRIES
PROSPECT_ONLY_INDUSTRIES.append([CLAY_PIT_NAME])

// Custom rule industries = have additional conditions in scpirt params
// Mostly means that if condition not fulfilled, industries will be omited
CUSTOM_RULE_INDUSTRIES <- [LUMBER_MILL_NAME, OIL_REF_NAME, FARM_NAME, FOREST_NAME, WATER_SUPPLY_NAME]

// Used to have separate build method, but it mostly merged with other methods
//SPECIAL_CHECK_INDUSTRIES <- [BANK_NAME, OIL_RIG_NAME, WATER_TOWER_NAME,	LUMBER_MILL_NAME]

///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

// Extend GS class
class IndustryConstructor extends GSController{
	BUILD_LIMIT = 0 											// Set from settings, in class constructor and each refresh.(initial is max ind per map, subs is max per refresh)
	CONTINUE_GAME_SCRIPT = null								// True if whole script must continue.
	INIT_PERFORMED = false										// True if IndustryConstructor.Init has run.
	FIRSTBUILD_PERFORMED = null								// True if IndustryConstructor.Load OR IndustryConstructor.BuildNewIndustry has run.
	BUILD_SPEED = 0											// Global build speed variable

	CLUSTERNODE_LIST_IND = [] 									// Sub-array of industry types, for cluster builder
	CLUSTERNODE_LIST_COUNT = []								// Sub-array of industry count, for cluster builder
	CLUSTERTILE_LIST = []										// Sub-array of node tiles, must be used with above 2D

	TOWNNODE_LIST_TOWN = []									// Sub-array of town ids registered
	TOWNNODE_LIST_IND = [] 									// Sub-array of industry types, for town builder
	TOWNNODE_LIST_COUNT = []									// Sub-array of industry count, for town builder

	// User variables
	DENSITY_IND_TOTAL = 0										// Set from settings, in IndustryConstructor.Init. Total industries, integer always >= 1
	DENSITY_IND_MIN = 0										// Set from settings, in IndustryConstructor. Init.Min industry density %, float always < 1.
	DENSITY_IND_MAX = 0										// Set from settings, in IndustryConstructor.Init. Max industry density %, float always > 1.
	DENSITY_RAW_PROP = 0										// Set from settings, in IndustryConstructor.Init. Primary industry proportion, float always < 1.
	DENSITY_PROC_PROP = 0										// Set from settings, in IndustryConstructor.Init. Secondary industry proportion, float always < 1.
	DENSITY_TERT_PROP = 0										// Set from settings, in IndustryConstructor.Init. Tertiary industry proportion, float always < 1.
	DENSITY_SPEC_PROP = 0										// Set from settings, in IndustryConstructor.Init. Special industry proportion, float always < 1.
	DENSITY_RAW_METHOD = 0										// Set from settings, in IndustryConstructor.Init.
	DENSITY_PROC_METHOD = 0									// Set from settings, in IndustryConstructor.Init.
	DENSITY_TERT_METHOD = 0									// Set from settings, in IndustryConstructor.Init.

	// supposed to contain last error during placement
	MSG_CLUSTER_CENTER_TASK = "Can't have cluster center here: "
	MSG_LAST_FAIL = ""
	
	
	//will be updated once
	MAP_SIZE_X = 1.0
	MAP_SIZE_Y = 1.0
	MAP_SCALE = 1.0
	IND_TYPE_COUNT = 0											// Count of industries in this.IND_TYPE_LIST, set in IndustryConstructor.Init.
	CARGO_PAXID = 0											// Passenger cargo ID, set in IndustryConstructor.Init.

	RAW_INDUSTRY_LIST = []										// Array of raw industry type ID's, set in IndustryConstructor.Init.
	SECONDARY_INDUSTRY_LIST = []								// Array of processor industry type ID's, set in IndustryConstructor.Init.
	TERTIARY_INDUSTRY_LIST = []								// Array of tertiary industry type ID's, set in IndustryConstructor.Init.
	SPECIAL_INDUSTRY_LIST = []									//Reserved

	// will be updated once on each run
	IS_CLIMATE_ARCTIC = false
	IS_CLIMATE_TROPIC = false
	IS_PROSPECT_ENABLED = false

	CUR_IND_NAME = ""
	CUR_IND_IS_MARINE = false
	CUR_IND_IS_CITY_ONLY = false
	CUR_IND_IS_NEAR_CITY = false
	CUR_IND_IS_COAST_ONLY = false
	CUR_IND_IS_TEMP_ONLY = false
	CUR_IND_IS_DESERT_ONLY = false
	CUR_IND_UNDER_CUSTOM_RULE = false
	CUR_IND_IS_PROSPECT_ONLY = false
	
	OPT_CUSTOM_SCALE = 0
	OPT_CLUSTER_MIN_TO_IND = 0
	OPT_CLUSTER_MIN_TO_CLUSTER = 0
	OPT_CLUSTER_MIN_TO_TOWN = 0
	OPT_CLUSTER_NODES = 0
	OPT_CLUSTER_MAX_RADIUS = 0
	OPT_CLUSTER_MIN_RADIUS = 0
	OPT_SCATTERED_MIN_TO_IND = 0
	OPT_SCATTERED_MIN_TO_TOWN = 0	
	OPT_TOWN_MAX_RADIUS = 0
	OPT_TOWN_MIN_IND = 0

	constructor(){
		INIT_PERFORMED = false
		CONTINUE_GAME_SCRIPT = true
		MAP_SIZE_X = GSMap.GetMapSizeX()
		MAP_SIZE_Y = GSMap.GetMapSizeY()
		BUILD_LIMIT = GSController.GetSetting("BUILD_LIMIT")
	}
}

function IndustryConstructor::Start(){
	LogDebugInfo("Starting Industry Constructor!")

	GSGame.Pause()
	this.Init()
	GSGame.Unpause()

	if (!FIRSTBUILD_PERFORMED){
		this.BuildNewIndustry()
		this.ErrorHandler()

		//GSNews.Create(GSNews.NT_GENERAL, GSText(GSText.END_START), GSCompany.COMPANY_INVALID)
	}
	SleepIfSlowBuild(Helper.Max(1, GSController.GetSetting("BUILD_SPEED") * 2252))


	/* this is supposed to be scenario-only script, but may be restored into ingame 
	local LOOP_START_TICK = 0
	local TICKS_USED = 0

	while (CONTINUE_GAME_SCRIPT){
		while(GSGame.IsPaused())
			SleepIfSlowBuild(100 * GSController.GetSetting("BUILD_SPEED"))

		// Get start tick
		LOOP_START_TICK = GSController.GetTick()

		// Assign build limit
		this.BUILD_LIMIT = GSController.GetSetting("BUILD_LIMIT")

		// Call build function, if manage is set in settings
		if (GSController.GetSetting("MANAGE_BOOL")){
			this.BuildNewIndustry()
			this.ErrorHandler()
		}

		// Call event handler function
		this.HandleEvents()
		this.ErrorHandler()

		// Pause for x months
		// - Assign used ticks
		TICKS_USED = GSController.GetTick() - LOOP_START_TICK
		// - 74 ticks = 1 day , 518 ticks = 1 week, 2252 ticks = 1 month, 27029 ticks = 1 year

		SleepIfSlowBuild(Helper.Max(1, (GSController.GetSetting("BUILD_SPEED") * 2252) - TICKS_USED))
	}*/

	LogDebugWarning("The game script has ended, and will now crash.")
}

function IndustryConstructor::Init(){
	if (!CONTINUE_GAME_SCRIPT)
		return

	LogDebugInfo("+==============================+")
	LogDebugInfo("Initializing...")

	// Set Advanced Setting parameters
	// - Check for multiple ind per town setting

	if (GSGameSettings.IsValid("multiple_industry_per_town")){
		if (GSGameSettings.GetValue("multiple_industry_per_town") == 0)
			LogDebugWarning("Multiple industries per town disabled, will slow down or prevent some build methods!")
		}
	else LogDebugError("Multiple industries per town setting could not be detected!")

	// Assign PAX cargo id
	// - Create cargo list
	local CARGO_LIST = GSCargoList()

	// - Loop for each cargo, Assign passenger cargo ID
	foreach (CARGO_ID in CARGO_LIST)
		if (GSCargo.GetTownEffect(CARGO_ID) == GSCargo.TE_PASSENGERS) CARGO_PAXID = CARGO_ID


	LogDebugInfo(">This is a new game, preparing...")

	// Check if there are industries on map (if user has not set to funding only error...)
	// Define new industry list
	local IND_LIST = GSIndustryList()
	// Count industries
	local IND_LIST_COUNT = IND_LIST.Count()

	if (IND_LIST_COUNT > 0)
		LogDebugWarning(">There are " + IND_LIST_COUNT + " industries on the map, when there must be none!")


	// If there are industries on the map
	/*if (false) {
		LogDebugWarning(">There are " + IND_LIST_COUNT + " industries on the map, when there must be none!")
		CONTINUE_GAME_SCRIPT = false
		return
	}*/


	local IND_TYPE_LIST = GSIndustryTypeList()
	local IND_TYPE_COUNT = IND_TYPE_LIST.Count()
	LogInfo(">There are " + IND_TYPE_COUNT + " industry types.")
	foreach(IND_ID, _ in IND_TYPE_LIST){
		local IND_NAME = GSIndustryType.GetName(IND_ID)

		/*foreach(SPECIAL_NAME in SPECIAL_CHECK_INDUSTRIES){
			if (IND_NAME == SPECIAL_NAME){
				LogInfo(" ~Special industry, will be prospected only on certain conditions: " + IND_NAME)
				SPECIAL_INDUSTRY_LIST.push(IND_ID)
				IS_SPECIAL = true
				break
			}
		}*/

		local IS_RAW = GSIndustryType.IsRawIndustry(IND_ID)
		local IS_PROC = GSIndustryType.IsProcessingIndustry(IND_ID)

		if (IS_RAW){
			LogInfo(" ~Raw Industry: " + IND_NAME)
			RAW_INDUSTRY_LIST.push(IND_ID)
		}
		
		if (IS_PROC){
			LogInfo(" ~Secondary Industry: " + IND_NAME)
			SECONDARY_INDUSTRY_LIST.push(IND_ID)
		}
		
		if (!IS_RAW && !IS_PROC){
				LogInfo(" ~Tertiary Industry: " + IND_NAME)
				TERTIARY_INDUSTRY_LIST.push(IND_ID)			
		}
	}
	
	local RAW_COUNT = RAW_INDUSTRY_LIST.len()
	local SECONDARY_COUNT = SECONDARY_INDUSTRY_LIST.len()
	local TERTIARY_COUNT = TERTIARY_INDUSTRY_LIST.len()
	local SPECIAL_COUNT = SPECIAL_INDUSTRY_LIST.len()

	LogInfo(">There are " + RAW_COUNT + " Primary industry types.")
	LogInfo(">There are " + SECONDARY_COUNT + " Secondary industry types.")
	LogInfo(">There are " + TERTIARY_COUNT + " Tertiary industry types.")
	LogInfo(">There are " + SPECIAL_COUNT + " Special industry types.")

	MAP_SCALE = sqrt(MAP_SIZE_X * MAP_SIZE_Y / (256.0 * 256.0))
	LogInfo(">Map size scale is: " + MAP_SCALE)

	local RAW_PROP = GSController.GetSetting("DENSITY_RAW_PROP").tofloat()
	local PROC_PROP = GSController.GetSetting("DENSITY_PROC_PROP").tofloat()
	local TERT_PROP = GSController.GetSetting("DENSITY_TERT_PROP").tofloat()
	local SPEC_PROP = GSController.GetSetting("DENSITY_SPEC_PROP").tofloat()		
	
	local TOTAL_PROP = RAW_PROP + PROC_PROP + TERT_PROP + SPEC_PROP
	DENSITY_RAW_PROP = RAW_PROP.tofloat() / TOTAL_PROP
	DENSITY_PROC_PROP = PROC_PROP.tofloat() / TOTAL_PROP
	DENSITY_TERT_PROP = TERT_PROP.tofloat() / TOTAL_PROP
	DENSITY_SPEC_PROP = SPEC_PROP.tofloat() / TOTAL_PROP
	
	DENSITY_IND_TOTAL = (GSController.GetSetting("DENSITY_IND_TOTAL") * MAP_SCALE).tointeger()
	DENSITY_IND_TOTAL = max(DENSITY_IND_TOTAL, 1)
	
	DENSITY_IND_MIN = GSController.GetSetting("DENSITY_IND_MIN").tofloat() / 100.0
	DENSITY_IND_MAX = GSController.GetSetting("DENSITY_IND_MAX").tofloat() / 100.0

	DENSITY_RAW_METHOD = GSController.GetSetting("DENSITY_RAW_METHOD")
	DENSITY_PROC_METHOD = GSController.GetSetting("DENSITY_PROC_METHOD")
	DENSITY_TERT_METHOD = GSController.GetSetting("DENSITY_TERT_METHOD")
	
	if (RAW_COUNT < 1) {RAW_PROP = 0.0; DENSITY_RAW_PROP = 0}
	if (SECONDARY_COUNT < 1) {PROC_PROP = 0.0; DENSITY_PROC_PROP = 0}
	if (TERTIARY_COUNT < 1) {TERT_PROP = 0.0; DENSITY_TERT_PROP = 0}
	if (SPECIAL_COUNT < 1) {SPEC_PROP = 0.0; DENSITY_SPEC_PROP = 0}
	
	// Not my idea, looks very bad
	if (GSGameSettings.IsValid("multiple_industry_per_town")){
		LogDebugWarning("Script may've changed your moultiple industry per town setting to match script's ")
		GSGameSettings.SetValue("multiple_industry_per_town", GSController.GetSetting("MULTI_IND_TOWN"))
	}
	
	if (GSGameSettings.IsValid("oil_refinery_limit")){
		GSGameSettings.SetValue("oil_refinery_limit", GSController.GetSetting("MAX_OIL_DIST"))
		LogDebugWarning("Script may've changed your oil ref limit setting to match script's ")
	} else {
		LogDebugError("Max distance from edge for Oil Refineries setting could not be detected!")
	}


	// - Display status msgs
	LogDebugInfo(">Total industries assigned: " + DENSITY_IND_TOTAL)
	LogDebugInfo(">Min per industry assigned: " + DENSITY_IND_MIN)
	LogDebugInfo(">Max per industry assigned: " + DENSITY_IND_MAX)
	LogDebugInfo(">Primary industry proportion assigned: " + DENSITY_RAW_PROP)
	LogDebugInfo(">Secondary industry proportion assigned: " + DENSITY_PROC_PROP)
	LogDebugInfo(">Tertiary industry proportion assigned: " + DENSITY_TERT_PROP)
	LogDebugInfo(">Special industry proportion assigned: " + DENSITY_SPEC_PROP)
	LogDebugInfo(">Primary industry method assigned: " + DENSITY_RAW_METHOD)
	LogDebugInfo(">Secondary industry method assigned: " + DENSITY_PROC_METHOD)
	LogDebugInfo(">Tertiary industry method assigned: " + DENSITY_TERT_METHOD)

	// Declare function status
	this.INIT_PERFORMED = true
}


function IndustryConstructor::RemoveUnbuildableFromList(LIST) {
	local FILTERED_LIST = []
	local SKIPPED_NAMES_LIST = []

	foreach(CURRENT_IND_ID in LIST){
		if (CanBuildIndustry(CURRENT_IND_ID))
			FILTERED_LIST.push(CURRENT_IND_ID)
		else
			SKIPPED_NAMES_LIST.push(GSIndustryType.GetName(CURRENT_IND_ID))
	}

	if (SKIPPED_NAMES_LIST.len() > 0){
		LogDebugWarning(" ~Skipping industry, probably because too early: ")
		foreach(NAME in SKIPPED_NAMES_LIST)
			LogDebugWarning(NAME)

		if (SKIPPED_NAMES_LIST.len() == LIST.len() && LIST.len() > 2)
			LogDebugError("Probably Newgrf prevents any building. Allow build in grf parameters and do not forget to turn if off after this script")
		}

	return FILTERED_LIST
}


function IndustryConstructor::TryToBuildIndustriesFromList(IND_LIST, IND_MIN, IND_MAX, CURRENT_METHOD_FUNC) {
	local ALL_TYPES_BUILT_COUNT = 0

	LogInfo("All target" + PadRight("", 40) + "Min/Max: " + IND_MIN + "/" + IND_MAX + "  " )

	foreach(CURRENT_IND_ID in IND_LIST){
		GatherIndustryInfo(CURRENT_IND_ID)

		local EXIST_IND_LIST = GSIndustryList()
		local EXIST_IND_COUNT = 0
		foreach(EXIST_IND_ID,_ in EXIST_IND_LIST){
			if (GSIndustry.GetIndustryType(EXIST_IND_ID) == CURRENT_IND_ID) 
				EXIST_IND_COUNT++
		}

		local stats = " ~Building: " + PadRight(CUR_IND_NAME, 28) + "Exists/Max: " + EXIST_IND_COUNT +"/"+ IND_MAX
		LogInfo(stats)

		local TARGET_IND_AMOUNT = IND_MIN  +  GSBase.RandRange(IND_MAX - IND_MIN) - EXIST_IND_COUNT
		TARGET_IND_AMOUNT = max(TARGET_IND_AMOUNT, 0)

		local CURRENT_BUILD_COUNT = 0
		if (TARGET_IND_AMOUNT > 0){
			for(local i = 0; i < TARGET_IND_AMOUNT; i++)
				if (TryToBuildIndustry(CURRENT_METHOD_FUNC, CURRENT_IND_ID))
					CURRENT_BUILD_COUNT++

			ALL_TYPES_BUILT_COUNT += CURRENT_BUILD_COUNT
		}

		local stats =  PadRight("", 50) + "Asked/Created : " + PadRight(TARGET_IND_AMOUNT, 2) + "/"+ PadRight(CURRENT_BUILD_COUNT, 2)
		if (TARGET_IND_AMOUNT > 0 && (CURRENT_BUILD_COUNT + EXIST_IND_COUNT) == 0)
			LogDebugError(stats)
		else
			LogInfo(stats)
	}
	return ALL_TYPES_BUILT_COUNT
}

// Cannot prospect != will not build
function IndustryConstructor::CanProspectAbnormal(CURRENT_IND_ID) {
	// don't even know why it was in script
	if (ArrContains([FARM_NAME, FOREST_NAME], CUR_IND_NAME))
		if (!IS_CLIMATE_ARCTIC)
			return false

	if (CUR_IND_NAME == WATER_SUPPLY_NAME)
		if (!IS_CLIMATE_TROPIC)
			return false

	return true
}

function IndustryConstructor::TryToBuildIndustry(CURRENT_METHOD_FUNC, CURRENT_IND_ID) {
	if (CUR_IND_UNDER_CUSTOM_RULE && !SpecialCheckBeforeBuild) {
		LogDebugWarning("Special rule not passed. Skip.")
		return false
	}

	if (IS_PROSPECT_ENABLED) {
		if ((CUR_IND_UNDER_CUSTOM_RULE && CanProspectAbnormal(CURRENT_IND_ID)) || CUR_IND_IS_COAST_ONLY || CUR_IND_IS_PROSPECT_ONLY) {
			LogDebugWarning("Prospecting is prioritised way to create this industry.")
			return GSIndustryType.ProspectIndustry (CURRENT_IND_ID)
		}
	}

	LogDebugInfo("Custom create method will be used, not prospecting")
	return CURRENT_METHOD_FUNC(CURRENT_IND_ID)
}

function IndustryConstructor::GetBuildMethod(CURRENT_METHOD) {
	if      (CURRENT_METHOD == 1)	return TownBuildMethod
	else if (CURRENT_METHOD == 2)	return ClusterBuildMethod
	else if (CURRENT_METHOD == 3)	return ScatteredBuildMethod
	else if (CURRENT_METHOD == 4)	return RandomBuildMethod
	else{ 
		LogDebugError(" ~IndustryConstructor.BuildSwitch: Incorrect build method chosen!")
		this.ErrorHandler()
		}
}

function IndustryConstructor::GatherMainInfo() {
	IS_CLIMATE_ARCTIC = (GSGame.GetLandscape () == GSGame.LT_ARCTIC)
	IS_CLIMATE_TROPIC = (GSGame.GetLandscape () == GSGame.LT_TROPIC)
	IS_PROSPECT_ENABLED = GSController.GetSetting("PROS_BOOL")

}

function IndustryConstructor::GatherIndustryInfo(CURRENT_IND_ID) {
	CUR_IND_NAME = GSIndustryType.GetName(CURRENT_IND_ID)
	CUR_IND_IS_MARINE = ArrContains(SEA_INDUSTRIES, CUR_IND_NAME)
	CUR_IND_IS_CITY_ONLY = ArrContains(CITY_INDUSTRIES, CUR_IND_NAME)
	CUR_IND_IS_NEAR_CITY = ArrContains(NEAR_CITY_INDUSTRIES, CUR_IND_NAME)
	CUR_IND_IS_COAST_ONLY = ArrContains(COAST_INDUSTRIES, CUR_IND_NAME)
	CUR_IND_IS_TEMP_ONLY = ArrContains(TEMP_ONLY_INDUSTRIES, CUR_IND_NAME)
	CUR_IND_IS_DESERT_ONLY = ArrContains(DESERT_ONLY_INDUSRIES, CUR_IND_NAME)
	CUR_IND_UNDER_CUSTOM_RULE = ArrContains(CUSTOM_RULE_INDUSTRIES, CUR_IND_NAME)
	CUR_IND_IS_PROSPECT_ONLY =  ArrContains(PROSPECT_ONLY_INDUSTRIES, CUR_IND_NAME)	

	LogDebugInfo("CUR_IND_NAME: " + CUR_IND_NAME)
	LogDebugInfo("CUR_IND_IS_MARINE: " + CUR_IND_IS_MARINE)
	LogDebugInfo("CUR_IND_IS_CITY_ONLY: " + CUR_IND_IS_CITY_ONLY)
	LogDebugInfo("CUR_IND_IS_NEAR_CITY: " + CUR_IND_IS_NEAR_CITY)
	LogDebugInfo("CUR_IND_IS_COAST_ONLY: " + CUR_IND_IS_COAST_ONLY)
	LogDebugInfo("CUR_IND_IS_TEMP_ONLY: " + CUR_IND_IS_TEMP_ONLY)
	LogDebugInfo("CUR_IND_IS_DESERT_ONLY: " + CUR_IND_IS_DESERT_ONLY)
	LogDebugInfo("CUR_IND_UNDER_CUSTOM_RULE: " + CUR_IND_UNDER_CUSTOM_RULE)
	LogDebugInfo("CUR_IND_IS_PROSPECT_ONLY: " + CUR_IND_IS_PROSPECT_ONLY)
	
	//Mostly for perfomance and cleanup of multiplier everywhere
	OPT_CLUSTER_NODES = GetSetting("CLUSTER_NODES")
	OPT_CUSTOM_SCALE = GetSetting("SOME_MULTIPLIER").tofloat() / 10
	
	OPT_CLUSTER_MIN_TO_IND = GetSetting("CLUSTER_MIN_IND") * OPT_CUSTOM_SCALE
	OPT_CLUSTER_MIN_TO_CLUSTER = GetSetting("CLUSTER_MIN_NODE") * OPT_CUSTOM_SCALE
	OPT_CLUSTER_MIN_TO_TOWN = GetSetting("CLUSTER_MIN_TOWN") * OPT_CUSTOM_SCALE	
	OPT_CLUSTER_MAX_RADIUS = GetSetting("CLUSTER_RADIUS_MAX") * OPT_CUSTOM_SCALE
	OPT_CLUSTER_MIN_RADIUS = GetSetting("CLUSTER_RADIUS_MIN") * OPT_CUSTOM_SCALE
	
	OPT_SCATTERED_MIN_TO_IND = GetSetting("SCATTERED_MIN_IND") * OPT_CUSTOM_SCALE
	OPT_SCATTERED_MIN_TO_TOWN = GetSetting("SCATTERED_MIN_TOWN") * OPT_CUSTOM_SCALE
	
	OPT_TOWN_MAX_RADIUS = GetSetting("TOWN_MAX_RADIUS") * OPT_CUSTOM_SCALE
	OPT_TOWN_MIN_IND = GetSetting("TOWN_MIN_IND") * OPT_CUSTOM_SCALE

}


//Function to call to build industries.
//Call to build initial and each refresh period
//Runs through each industry class, checks if it has built that class already, builds random class once if not and sets to true,
//     if all have been built resets all to false up until global max per refresh
//Builds a global set proportion of each industry class, minimum 1 of each industry.
//Checks how many of each industry are existing (if 0 then has a chance to build 1).
//Builds up to a random number of each industry, between min and max according to proportion.
//Builds each industry according to build method.
//Accounts for special industry types.

function IndustryConstructor::BuildNewIndustry() {

	if (!CONTINUE_GAME_SCRIPT)
		return

	LogDebugInfo("+==============================+")
	LogDebugInfo("Building industries...")


	if (!this.INIT_PERFORMED){
		LogDebugError(">IndustryConstructor.Build: Script uninitialized!")
		return
	}

	// If first build has not happened, then this is a new game
	local LOCAL_BUILD_LIMIT = 0			// Build limit in this function
	local LOCAL_IND_MAX = 0.0				// Max factor in this function
	if (!FIRSTBUILD_PERFORMED){
		LOCAL_BUILD_LIMIT = max(IND_TYPE_COUNT, DENSITY_IND_TOTAL)
		LOCAL_IND_MAX = 1.0

		LogDebugInfo(">This is a new game, build limit is: " + LOCAL_BUILD_LIMIT)

		// Set first build
		FIRSTBUILD_PERFORMED = true
	}
	// Else build has happened
	else{
		// Set build limit to a random up to global
		LOCAL_BUILD_LIMIT = GSBase.RandRange(BUILD_LIMIT + 1)

		// Set max factor to global
		LOCAL_IND_MAX = DENSITY_IND_MAX

		LogDebugInfo(">This is a continuing game, build limit is: " + LOCAL_BUILD_LIMIT + " / " + BUILD_LIMIT + " (Max)")
	}

	GatherMainInfo()

	local AVALIABLE_RAW_INDUSTRY_LIST = RemoveUnbuildableFromList(RAW_INDUSTRY_LIST)
	local AVALIABLE_SECONDARY_INDUSTRY_LIST = RemoveUnbuildableFromList(SECONDARY_INDUSTRY_LIST)
	local AVALIABLE_TERTIARY_INDUSTRY_LIST = RemoveUnbuildableFromList(TERTIARY_INDUSTRY_LIST)
	local AVALIABLE_SPECIAL_INDUSTRY_LIST = RemoveUnbuildableFromList(SPECIAL_INDUSTRY_LIST)

	// This al is terrible and needs rewrites (count target from the start of script without all mess), but will do for now
	local PRIMARY_PERFORMED = false
	local SECONDARY_PERFORMED = false
	local TERTIARY_PERFORMED = false
	local SPECIAL_PERFORMED = false
	
	if (AVALIABLE_RAW_INDUSTRY_LIST.len() < 1) PRIMARY_PERFORMED = true
	if (AVALIABLE_SECONDARY_INDUSTRY_LIST.len() < 1) SECONDARY_PERFORMED = true
	if (AVALIABLE_TERTIARY_INDUSTRY_LIST.len() < 1) TERTIARY_PERFORMED = true
	if (AVALIABLE_SPECIAL_INDUSTRY_LIST.len() < 1) SPECIAL_PERFORMED = true

	// Loop until build limit is reached
	// Total build count for each build function
	local BUILT_COUNT = 0					
	while(LOCAL_BUILD_LIMIT > BUILT_COUNT){
		// Selects a random industry class, not yet built....GSBase.RandRange(Max - 1)
		// - Build "town" first
		local NEXT_BUILD_CLASS = GSBase.RandRange(4)
		if (!PRIMARY_PERFORMED && DENSITY_RAW_METHOD == 1) NEXT_BUILD_CLASS = 0
		if (!SECONDARY_PERFORMED && DENSITY_PROC_METHOD == 1) NEXT_BUILD_CLASS = 1
		if (!TERTIARY_PERFORMED && DENSITY_RAW_METHOD == 1) NEXT_BUILD_CLASS = 2

		local CURRENT_METHOD = -1 // Current build method
		local CURRENT_IND_PROP = null	// Current industry proportion for each loop
		local CURRENT_LIST = null		// Current industry list for each loop

		switch(NEXT_BUILD_CLASS)
		{
			case 0:
				if (PRIMARY_PERFORMED) continue
				LogDebugInfo("+------------------------------+")
				LogDebugInfo(">Building primary industries")
				CURRENT_LIST = AVALIABLE_RAW_INDUSTRY_LIST
				CURRENT_METHOD = DENSITY_RAW_METHOD
				CURRENT_IND_PROP = DENSITY_RAW_PROP
				PRIMARY_PERFORMED = true
				break
			case 1:
				if (SECONDARY_PERFORMED) continue
				LogDebugInfo("+------------------------------+")
				LogDebugInfo(">Building secondary industries")
				CURRENT_LIST = AVALIABLE_SECONDARY_INDUSTRY_LIST
				CURRENT_METHOD = DENSITY_PROC_METHOD
				CURRENT_IND_PROP = DENSITY_PROC_PROP
				SECONDARY_PERFORMED = true
				break
			case 2:
				if (TERTIARY_PERFORMED) continue
				LogDebugInfo("+------------------------------+")
				LogDebugInfo("Building tertiary industries")
				CURRENT_LIST = AVALIABLE_TERTIARY_INDUSTRY_LIST
				CURRENT_METHOD = DENSITY_TERT_METHOD
				CURRENT_IND_PROP = DENSITY_TERT_PROP
				TERTIARY_PERFORMED = true
				break
			case 3:
				if (SPECIAL_PERFORMED) continue
				LogDebugInfo("+------------------------------+")
				LogDebugInfo(">Building special industries")
				CURRENT_LIST = AVALIABLE_SPECIAL_INDUSTRY_LIST
				CURRENT_METHOD = 5
				CURRENT_IND_PROP = DENSITY_SPEC_PROP
				SPECIAL_PERFORMED = true
				break
			default:
				LogDebugError(">IndustryConstructor.BuildNewIndustry: Incorrect industry class chosen!")
				return
		}

		LogDebugInfo(" ~Method: " + CURRENT_METHOD)
		local CURRENT_METHOD_FUNC = GetBuildMethod(CURRENT_METHOD)
		local CURRENT_LIST_COUNT = CURRENT_LIST.len()

		// Set densities
		// Mins are how much a map will start with and the start count for building extra.
		// Maxes are how much a map could get based on chance. At least 1 per industry

		local CLASS_MAX = (DENSITY_IND_TOTAL * CURRENT_IND_PROP * LOCAL_IND_MAX).tointeger()
		CLASS_MAX = max(CURRENT_LIST_COUNT, CLASS_MAX)
		LogDebugInfo(" ~Industry class maximum is: " + CLASS_MAX)

		local CLASS_MIN = (CLASS_MAX * DENSITY_IND_MIN).tointeger()
		LogDebugInfo(" ~Industry class minimum is: " + CLASS_MIN)


		local IND_MIN = (CLASS_MIN * 1.0 / CURRENT_LIST_COUNT + 1).tointeger()
		LogDebugInfo(" ~Industry type minimum is: " + IND_MIN)

		local IND_MAX = (CLASS_MAX * 1.0 / CURRENT_LIST_COUNT + 1).tointeger()
		LogDebugInfo(" ~Industry type maximum is: " + IND_MAX)

		BUILT_COUNT += TryToBuildIndustriesFromList(CURRENT_LIST, IND_MIN, IND_MAX, CURRENT_METHOD_FUNC)

		LogDebugInfo(">Build a current total of: " + BUILT_COUNT + " / " + LOCAL_BUILD_LIMIT)

		// Reset randoms if necessary
		if (PRIMARY_PERFORMED && SECONDARY_PERFORMED && TERTIARY_PERFORMED && SPECIAL_PERFORMED){
			PRIMARY_PERFORMED = false
			SECONDARY_PERFORMED = false
			TERTIARY_PERFORMED = false
			SPECIAL_PERFORMED = false

			LogDebugInfo("+------------------------------+")
			LogDebugInfo(">Built all industry class types ")
			return
		}
	}
}

function IndustryConstructor::SpecialCheckBeforeBuild(){
	// Switch ind id for each type, must be same as in CUSTOM_RULE_INDUSTRIES
	switch(CUR_IND_NAME){

		case BANK_NAME:
			// Check towns with pop > parameter
			local LOCAL_TOWN_LIST = GSTownList()
			LOCAL_TOWN_LIST.Valuate(GSTown.GetPopulation)
			LOCAL_TOWN_LIST.RemoveBelowValue(GSController.GetSetting("SPEC_BANK_MINPOP"))

			// Check if valid
			if (LOCAL_TOWN_LIST.IsEmpty()){
				LogDebugWarning(" ~IndustryConstructor.SpecialBuildMethod: No towns with more than " + GSController.GetSetting("SPEC_BANK_MINPOP") + " for Banks!")
				return false
			}
			break

		// Check if current date is before param
		case OIL_RIG_NAME:
			if (GSDate.GetCurrentDate() < GSDate.GetDate(GSController.GetSetting("SPEC_RIG_MINYEAR"), 1, 1)){
				LogDebugWarning(" ~IndustryConstructor.SpecialBuildMethod: Year is less than " + GSController.GetSetting("SPEC_RIG_MINYEAR") + " for Oil Rig!")
				return false
			}
			break

		// Check towns with pop > parameter
		case WATER_TOWER_NAME:
			local LOCAL_TOWN_LIST = GSTownList()
			LOCAL_TOWN_LIST.Valuate(GSTown.GetPopulation)
			LOCAL_TOWN_LIST.RemoveBelowValue(GSController.GetSetting("SPEC_WTR_MINPOP"))

			if (LOCAL_TOWN_LIST.IsEmpty()){
				LogDebugWarning(" ~IndustryConstructor.SpecialBuildMethod: No towns with more than " + GSController.GetSetting("SPEC_WTR_MINPOP") + " for Water Towers!")
				return false
			}
			break

		case LUMBER_MILL_NAME:
			if (!GSController.GetSetting("SPEC_LBR_BOOL"))
				return false
			break

		default:
			LogDebugError(" ~IndustryConstructor.SpecialBuildMethod: Industry " + CUR_IND_NAME + " not supported!")
	}

	return true
}

function IndustryConstructor::SampleWaterToLandRatio(NODE_TILE) {
	local RANDOM_JUMPER = RandomJumper()
	RANDOM_JUMPER.Start(NODE_TILE, 10, OPT_CLUSTER_MIN_RADIUS + 10)
	
	local TOTAL_SEA = 0
	local SAMPLE_COUNT = 20
	for (local i = 0; i < SAMPLE_COUNT; i++){
		local TILE_ID = RANDOM_JUMPER.Jump()
		if (GSTile.IsWaterTile(TILE_ID))
			TOTAL_SEA++
	}
	
	return TOTAL_SEA.tofloat() / SAMPLE_COUNT
}

function IndustryConstructor::CanHaveClusterCenterOnTile(NODE_TILE) {
	if (GSMap.DistanceFromEdge(NODE_TILE) < OPT_CLUSTER_MIN_RADIUS){
		MSG_LAST_FAIL = "Edge distance failed: " + GSMap.DistanceFromEdge(NODE_TILE) + "   " + OPT_CLUSTER_MIN_RADIUS
		return false
	}
	
	local WATER_TO_LAND_RATIO = SampleWaterToLandRatio(NODE_TILE)
	if (CUR_IND_IS_MARINE){
		if (WATER_TO_LAND_RATIO == 1 || WATER_TO_LAND_RATIO < 0.2){
			MSG_LAST_FAIL = "Incorrect water ratio: " + WATER_TO_LAND_RATIO
			return false
		}
	} else {
		if (WATER_TO_LAND_RATIO > 0.4){
			MSG_LAST_FAIL = "Incorrect water ratio: " + WATER_TO_LAND_RATIO
			return false
		}
	}
	

	if (!CUR_IND_IS_MARINE){
		local TOWN_TILE = GSTile.GetClosestTown(NODE_TILE)
		if (TOWN_TILE != null) {			
			local CUR_DIST = GSTown.GetDistanceManhattanToTile(TOWN_TILE, NODE_TILE)
			if (CUR_DIST < OPT_CLUSTER_MIN_TO_TOWN){
				MSG_LAST_FAIL = "Cluster min distance to town failed"
				return false
			}
		}
	}

	local IND = this.GetClosestIndustry(NODE_TILE)
	if (IND != null){
		local CUR_DIST = GSIndustry.GetDistanceManhattanToTile(IND, NODE_TILE)
		if (CUR_DIST < OPT_CLUSTER_MIN_TO_IND) {
			MSG_LAST_FAIL = "Cluster min distance to any industry failed"
			return false
		}
	}

	if (CLUSTERNODE_LIST_IND.len() > 0)		
		for(local i = 0; i < CLUSTERTILE_LIST.len(); i++)
			if (GSTile.GetDistanceManhattanToTile(NODE_TILE, CLUSTERTILE_LIST[i]) < OPT_CLUSTER_MIN_TO_CLUSTER) {
				MSG_LAST_FAIL = "Cluster min distance to other clusters failed"
				return false
				}
	
	return true
}


function IndustryConstructor::CanBuildIndustryOnTile(TILE_ID) {
	local IS_WATER = GSTile.IsWaterTile(TILE_ID)
	if (CUR_IND_IS_MARINE) {
		local NEED_EDGE_DISTANCE = 15
		
		if (CUR_IND_NAME == OIL_REF_NAME)
			if (GSGameSettings.IsValid("oil_refinery_limit"))
				NEED_EDGE_DISTANCE = GSGameSettings.GetValue("oil_refinery_limit")
			
		if (GSMap.DistanceFromEdge(TILE_ID) < NEED_EDGE_DISTANCE){
			MSG_LAST_FAIL = "Distance to the edge"
			return false
		}
		
		if (!IS_WATER) MSG_LAST_FAIL = "Sea industree on land";
		return IS_WATER
	}
			
	local IS_BUILDABLE = GSTile.IsBuildable(TILE_ID)
	if (!IS_BUILDABLE){
		MSG_LAST_FAIL = "Tile has no buildable rectangle"
		return false
	}

	local TILE_TERRAIN = GSTile.GetTerrainType(TILE_ID)
	local IS_SNOW = (TILE_TERRAIN == GSTile.TERRAIN_SNOW)
	local IS_DESERT = (TILE_TERRAIN == GSTile.TERRAIN_DESERT)
	local IS_TEMP = (TILE_TERRAIN == GSTile.TERRAIN_NORMAL) || (TILE_TERRAIN == GSTile.TERRAIN_RAINFOREST)

	if (CUR_IND_IS_DESERT_ONLY && !IS_DESERT){
		MSG_LAST_FAIL = "Desert industry not in desert"
		return false
	}
	if (CUR_IND_IS_TEMP_ONLY && !IS_TEMP){
		MSG_LAST_FAIL = "Temp industry not in temp"
		return false
	}

	local IS_COAST = GSTile.IsCoastTile (TILE_ID)
	if (CUR_IND_IS_COAST_ONLY && !IS_COAST){
		MSG_LAST_FAIL = "Coast industry not on coast"
		return false
	}

	return true
}


// ONLY STORES 1 IND ID IN REFERENCE ARRAY, SO WILL ONLY SET EACH ITERATION TO CURRENT ID
// FINE ON MAP CREATION BUT WILL NOT WORK ON CONSECUTIVE RUNS
// TO FIX NEED A 2D ARRAY FOR THAT LIST< THEN TO LOOP FOR EACH IND ID IF HIT ON TOWN

function IndustryConstructor::TownBuildMethod(INDUSRTY_ID){
	// Create town list for townbuilder
	local LOCAL_TOWN_LIST = GSTownList()
	// Valuate by population
	LOCAL_TOWN_LIST.Valuate(GSTown.GetPopulation)
	// Remove below parameter
	LOCAL_TOWN_LIST.RemoveBelowValue(GSController.GetSetting("TOWN_MIN_POP"))

	// Check abnormal industries, for towns
	if (CUR_IND_NAME == OIL_REF_NAME){
		// - Check oil ind setting, and remove towns further from edge
		//   Mainly for speed purposes...
		if (GSGameSettings.IsValid("oil_refinery_limit")){
			// Valuate by edge distance
			LOCAL_TOWN_LIST.Valuate(GetTownDistFromEdge)
			// Remove towns farther than max, including town radius
			local MAX_DIST = GSGameSettings.GetValue("oil_refinery_limit") + OPT_TOWN_MAX_RADIUS - 2
			MAX_DIST = max(MAX_DIST, 0)
			LOCAL_TOWN_LIST.RemoveAboveValue(MAX_DIST)
		}
	}

	// Loop through town list arrays
	for(local i = 0; i < this.TOWNNODE_LIST_TOWN.len(); i++){

		// Remove towns with ind ID in if TOWN_MULTI_BOOL
		// If - TOWN_MULTI_BOOL
		if (GSController.GetSetting("TOWN_MULTI_BOOL") == 0){
			// - If current list id == ind id, remove town
			if (this.TOWNNODE_LIST_IND[i] == INDUSRTY_ID) LOCAL_TOWN_LIST.RemoveItem(this.TOWNNODE_LIST_TOWN[i])
		}

		// Remove towns with max
		// - If loop town count >= setting, remove from list
		if (this.TOWNNODE_LIST_COUNT[i] >= GSController.GetSetting("TOWN_MAX_IND")) LOCAL_TOWN_LIST.RemoveItem(this.TOWNNODE_LIST_TOWN[i])
	}

	// Check if the list is not empty
	if (LOCAL_TOWN_LIST.IsEmpty()){
		LogDebugError(" ~IndustryConstructor.TownBuildMethod: Town list is empty!")
		return false
	}

	local MAX_ATTEMPTS = LOCAL_TOWN_LIST.Count() * 3
	// Loop until tries are maxed (mainly debug)
	for (local CUR_TRY = 0; CUR_TRY < MAX_ATTEMPTS; CUR_TRY++){
		// Get start ID
		local TOWN_ID = LOCAL_TOWN_LIST.Begin()
		// Get random ID
		for(local i = 0; i < GSBase.RandRange(LOCAL_TOWN_LIST.Count()); i++){
			TOWN_ID = LOCAL_TOWN_LIST.Next()
		}

		LogDebugInfo("   ~Trying to build " + CUR_IND_NAME + " in " + GSTown.GetName(TOWN_ID))
		local TOWN_RADIUS = sqrt(GSTown.GetHouseCount(TOWN_ID).tofloat())

		local RANDOM_JUMPER = RandomJumper()
		local CITY_TILE = GSTown.GetLocation(TOWN_ID)
		RANDOM_JUMPER.Start(CITY_TILE, TOWN_RADIUS, OPT_TOWN_MAX_RADIUS)

		local DYNAMIC_TOWN_MIN_IND = OPT_TOWN_MIN_IND

		local BORDER_TILE = null
		local IND = null
		local DIST_TO_IND = 0
		for (local i = 0; i < 100; i++){
			BORDER_TILE = RANDOM_JUMPER.Jump()

			if (!GSMap.IsValidTile(BORDER_TILE))
				continue

			if (GSController.GetSetting("log_level") >= 4) 
				GSSign.BuildSign(BORDER_TILE, "Try")

			if (!CanBuildIndustryOnTile(BORDER_TILE))
				continue


			IND = this.GetClosestIndustry(BORDER_TILE)
			// - If not null (null - no indusrties)
			if (IND != null){
				// - Get distance
				DIST_TO_IND = GSIndustry.GetDistanceManhattanToTile(IND, BORDER_TILE)
				// - If less than minimum, re loop
				if (DIST_TO_IND < DYNAMIC_TOWN_MIN_IND)
					{
					DYNAMIC_TOWN_MIN_IND = DecToOne(DYNAMIC_TOWN_MIN_IND)
					continue
					}
			}

			if (GSIndustryType.BuildIndustry(INDUSRTY_ID, BORDER_TILE)){
				LogDebugInfo("   ~Built!")

				// Loop through town list arrays
				local EXIST_TOWN = false
				for(local i = 0; i < TOWNNODE_LIST_TOWN.len(); i++){
					// If current town is in array
					if (TOWNNODE_LIST_TOWN[i] == TOWN_ID){
						// Set bool
						EXIST_TOWN = true
						// Inc count in array
						TOWNNODE_LIST_COUNT[i]++
						// Set ind in array
						TOWNNODE_LIST_IND[i] = INDUSRTY_ID
					}
				}

				// If town was not in array
				if (!EXIST_TOWN){
					// Add town to array
					TOWNNODE_LIST_TOWN.push(TOWN_ID)
					// Add ind id to array
					TOWNNODE_LIST_IND.push(INDUSRTY_ID)
					// Add count to array
					TOWNNODE_LIST_COUNT.push(1)
				}
				return true
			}
		}
		LogIfDynamicChanged("TOWN_MIN_IND", DYNAMIC_TOWN_MIN_IND)
	}

	LogDebugError("IndustryConstructor.TownBuildMethod: Couldn't find a valid tile to set node on!")
	return false
}

function IndustryConstructor::LogAttemptFailure(CUR_ATTEMPT, MAX_ATTEMPTS, TASK_DESC, REASON_MSG){
	local MSG_LVL3 = false
	
	if (CUR_ATTEMPT > MAX_ATTEMPTS / 2)
		if (CUR_ATTEMPT % 10 == 0)
			MSG_LVL3 = true

	if (CUR_ATTEMPT > MAX_ATTEMPTS  - 10)
		MSG_LVL3 = true
	
	local stat = PadRight(CUR_ATTEMPT + "/" + MAX_ATTEMPTS, 10)
	local msg = stat +  CUR_IND_NAME + " " + TASK_DESC + REASON_MSG
	
	if (MSG_LVL3)
		LogInfo(msg)
	else
		LogDebugInfo(msg)
}

function IndustryConstructor::LogClusterAttemptFailure(CUR_ATTEMPT, MAX_ATTEMPTS){
	LogAttemptFailure(CUR_ATTEMPT, MAX_ATTEMPTS, MSG_CLUSTER_CENTER_TASK, MSG_LAST_FAIL)
	MSG_LAST_FAIL = ""
}

function IndustryConstructor::ClusterBuildMethod(INDUSRTY_ID){
	local LIST_VALUE = 0											// The point on the list surrently, to synchronise between lists
	local CLUSTER_CENTER_TILE = null
	local IND = null
	local DIST_TO_IND = 0
	
	MSG_LAST_FAIL = ""

	// Find if node exists already
	// - Check if node list has entries
	LogDebugInfo("Node list length: " + CLUSTERNODE_LIST_IND.len())
	if (CLUSTERNODE_LIST_IND.len() > 0){
		// - Loop through node list
		for(local i = 0; i < CLUSTERNODE_LIST_IND.len(); i++){
			// - If node for industry exists
			if (CLUSTERNODE_LIST_IND[i] == INDUSRTY_ID) {
				// - If industry count is less than or equal to global, node tile is that tile
				if (CLUSTERNODE_LIST_COUNT[i] < GSController.GetSetting("CLUSTER_NODES")){
					// Set node tile
					CLUSTER_CENTER_TILE = CLUSTERTILE_LIST[i]

					LogDebugInfo("Using existing node: " + GSMap.IsValidTile(CLUSTER_CENTER_TILE))
					// Inc list counter
					LIST_VALUE = i
					break
				}
			}
		}
	}
	// If node tile doesn't exist then get one
	if (CLUSTER_CENTER_TILE == null)
	{
		LogDebugInfo("Finding new node")
		local MAX_SEARCH_TRIES = 32 * 32 * MAP_SCALE

		for (local CUR_TRY = 0; CUR_TRY < MAX_SEARCH_TRIES; CUR_TRY++){
			WarningOnTryLimit(CUR_TRY, MAX_SEARCH_TRIES)

			CLUSTER_CENTER_TILE = Tile.GetRandomTile()

			if (!CanHaveClusterCenterOnTile(CLUSTER_CENTER_TILE)){
				LogClusterAttemptFailure(CUR_TRY, MAX_SEARCH_TRIES)
				continue
			}

			if (!CanBuildIndustryOnTile(CLUSTER_CENTER_TILE)){
				LogClusterAttemptFailure(CUR_TRY, MAX_SEARCH_TRIES)				
				continue
			}

			//LogDebugInfo("node fine")
			break
		}

		// Add to node list
		CLUSTERNODE_LIST_IND.push(INDUSRTY_ID)
		CLUSTERNODE_LIST_COUNT.push(0)
		CLUSTERTILE_LIST.push(CLUSTER_CENTER_TILE)
		LIST_VALUE = CLUSTERTILE_LIST.len() - 1

	}
	// Get tile to build industry on
	local TILE_ID = null


	// Debug sign
  	if (GSController.GetSetting("log_level") >= 4)
		GSSign.BuildSign(CLUSTER_CENTER_TILE,"Node tile: " + CUR_IND_NAME)

	local RANDOM_JUMPER =  RandomJumper()
	RANDOM_JUMPER.Start(CLUSTER_CENTER_TILE, 1, OPT_CLUSTER_MAX_RADIUS)

	local DYNAMIC_MIN = OPT_CLUSTER_MIN_RADIUS

	local MAX_ATTEMPTS = 100
	// Loop till built
	for (local CUR_TRY = 0; CUR_TRY < MAX_ATTEMPTS; CUR_TRY++){

		TILE_ID = RANDOM_JUMPER.Jump()

		// Check dist from ind
		// - Get industry
		IND = this.GetClosestIndustry(TILE_ID)
		// - If not null (null - no indusrties)
		if (IND != null){
			// - Get distance
			DIST_TO_IND = GSIndustry.GetDistanceManhattanToTile(IND, TILE_ID)

			// - If less than minimum, re loop
			if (DIST_TO_IND < DYNAMIC_MIN){
				if (DYNAMIC_MIN > 1) DYNAMIC_MIN--
				continue
			}
			// - If more than maximum, re loop
			//if (DIST_TO_IND > OPT_CLUSTER_MAX_RADIUS) continue
		}

		if (GSIndustryType.BuildIndustry(INDUSRTY_ID, TILE_ID)){
			CLUSTERNODE_LIST_COUNT[LIST_VALUE]++
			LogDebugInfo("Walk done")
			return true
		}

		WarningOnTryLimit(CUR_TRY, MAX_ATTEMPTS)
	}

	LogDebugError("IndustryConstructor.ClusterBuildMethod: Build failed!  " + CUR_IND_NAME)
	return false
}

function IndustryConstructor::ScatteredBuildMethod(INDUSRTY_ID){
	local TILE_ID = null
	local DIST_TO_TOWN = 0
	local IND = null
	local DIST_TO_IND = 0
	local DYNAMIC_SCATTERED_MIN_IND = OPT_SCATTERED_MIN_TO_IND
	local DYNAMIC_SCATTERED_MIN_TOWN = OPT_SCATTERED_MIN_TO_TOWN

	local MAX_ATTEMPTS = 200
	if (CUR_IND_IS_CITY_ONLY || CUR_IND_IS_NEAR_CITY)
		MAX_ATTEMPTS = 300

	for (local CUR_TRY = 0; CUR_TRY < MAX_ATTEMPTS; CUR_TRY++){
		TILE_ID = Tile.GetRandomTile()

		if (!CanBuildIndustryOnTile(TILE_ID))
			continue
		
		//Check distance from town		
		DIST_TO_TOWN = GSTown.GetDistanceManhattanToTile(GSTile.GetClosestTown(TILE_ID),TILE_ID)
		if (!CUR_IND_IS_CITY_ONLY && !CUR_IND_IS_NEAR_CITY)
			if (DIST_TO_TOWN < DYNAMIC_SCATTERED_MIN_TOWN){
				DYNAMIC_SCATTERED_MIN_TOWN = DecToOne(DYNAMIC_SCATTERED_MIN_TOWN)
				LogIfDynamicChanged("SCATTERED_MIN_TOWN", DYNAMIC_SCATTERED_MIN_TOWN)
				continue
				}			
		

		// Check dist from other industries may became problem for city industries if there is too few cities
		IND = this.GetClosestIndustry(TILE_ID)
		if (IND != null){
			DIST_TO_IND = GSIndustry.GetDistanceManhattanToTile(IND, TILE_ID)
			if (DIST_TO_IND < DYNAMIC_SCATTERED_MIN_IND) {
				DYNAMIC_SCATTERED_MIN_IND = DecToOne(DYNAMIC_SCATTERED_MIN_IND)
				LogIfDynamicChanged("SCATTERED_MIN_IND", DYNAMIC_SCATTERED_MIN_IND)
				continue
			}
		}

		if (GSIndustryType.BuildIndustry(INDUSRTY_ID, TILE_ID))
			return true

		WarningOnTryLimit(CUR_TRY, MAX_ATTEMPTS)
	}
	LogDebugError("IndustryConstructor.ScatteredBuildMethod: Build failed!  " + CUR_IND_NAME)
	return false
}

function IndustryConstructor::WarningOnTryLimit(CUR_TRY, MAX_ATTEMPTS){
	if (CUR_TRY > 0 && CUR_TRY % (MAX_ATTEMPTS / 3) == 0 )
		LogDebugInfo(" ~Tries left: " + (MAX_ATTEMPTS - CUR_TRY))
	//if (CUR_TRY == 0){
	//	LogDebugError("IndustryConstructor.RandomBuildMethod: Couldn't find a valid tile!")
}

function IndustryConstructor::RandomBuildMethod(INDUSRTY_ID){
	local TILE_ID = null
	local MAX_ATTEMPTS = 32 * 32 * MAP_SCALE

	for (local CUR_TRY = 0; CUR_TRY < MAX_ATTEMPTS; CUR_TRY++){
		// Get a random tile
		TILE_ID = Tile.GetRandomTile()

		if (!CanBuildIndustryOnTile(TILE_ID))
			continue

		if (GSIndustryType.BuildIndustry(INDUSRTY_ID, TILE_ID)) 
			return true

		WarningOnTryLimit(CUR_TRY, MAX_ATTEMPTS)
	}
	LogDebugError("IndustryConstructor.RandomBuildMethod: Build failed!  " + CUR_IND_NAME)
	return false
}

// NOTE: Has to be called from the HandleEvents of main.
function IndustryConstructor::HandleEvents(){

	if (!CONTINUE_GAME_SCRIPT)
		return

	LogDebugInfo("+==============================+")
	LogDebugInfo("Event handling...")

	while (GSEventController.IsEventWaiting()) {

  		local NEXT_EVENT = GSEventController.GetNextEvent()
  		switch (NEXT_EVENT.GetEventType()) {

		    case GSEvent.ET_INDUSTRY_OPEN:
		    	LogDebugInfo(">New industry opened event")

		      	local EVENT_CONTROLER = GSEventIndustryOpen.Convert(NEXT_EVENT)
		      	local INDUSTRY_ID  = EVENT_CONTROLER.GetIndustryID()		      
		      	local TILE_ID = GSIndustry.GetLocation(INDUSTRY_ID)
		      	// Demolish the industry
		      	GSTile.DemolishTile(TILE_ID)
		      	break

		    case GSEvent.ET_INDUSTRY_CLOSE:
		    
		    	LogDebugInfo(">Industry closed event")

		      	local EVENT_CONTROLER = GSEventIndustryClose.Convert(NEXT_EVENT)
		      	local INDUSTRY_ID  = EVENT_CONTROLER.GetIndustryID()
		      	// Do nothing, as reduced number of industries will be handeled in the next refresh loop
		      	break

		    default:		    
		    	LogDebugInfo(">Unhandled event")
		      	break
  		}
	}
}

// Custom get closest industry function
function IndustryConstructor::GetClosestIndustry(TILE){
	// Create a list of all industries
	local IND_LIST = GSIndustryList()

	if (IND_LIST.Count() == 0) return null

	// Valuate by distance from tile
	IND_LIST.Valuate(GSIndustry.GetDistanceManhattanToTile, TILE)

	// Sort smallest to largest
	IND_LIST.Sort(GSList.SORT_BY_VALUE, GSList.SORT_ASCENDING)

	// Return the top one
	return IND_LIST.Begin()
}

/*
//import("util.MinchinWeb", "MinchinWeb", 9)
//SpiralWalker <- Spiral.Walker.nut
require("Spiral.Walker.nut")
SpiralWalker <- _MinchinWeb_SW_

// Town house list function, returns a list of tiles with houses on from a town
function IndustryConstructor:: GetTownHouseList(TOWN_ID, CARGO_ID){
  	// Below requires
	//	TOWN_ID // ID of town
	//	CARGO_PAXID	//	Passenger cargo ID

	// Configure variables
	local HOUSE_COUNT_FACTOR = 1.25		// Account for multi tile buildings
	local MAX_TRIES = (128 * 128)			// Maximum size for search, to prevent infinite loop

	// Create a blank tile list
	local TOWN_HOUSE_LIST = GSTileList()
	local TOWN_HOUSE_LIST_COUNT = 0

	// Create a cargo counter
	local CARGO_COUNTER = 0
	//GSLog.Info(GSCargo.GetCargoLabel(CARGO_PAXID))

	// Get town house count
	local TOWN_HOUSE_COUNT = GSTown.GetHouseCount(TOWN_ID)

	// Set current tile
	local CURRENT_TILE = GSTown.GetLocation(TOWN_ID)

	//GSLog.Info(TOWN_ID)
	//GSLog.Info("===")
	//GSLogDebugInfo(GSTile.IsWithinTownInfluence(CURRENT_TILE)
	//GSLog.Info(GSTile.GetTownAuthority(CURRENT_TILE)) //= TOWN_ID
	//GSLog.Info(GSTile.IsBuildable(CURRENT_TILE))	//=false
	//GSLog.Info(GSTile.GetOwner(CURRENT_TILE))	//=-1
	//GSLog.Info(GSTile.IsStationTile(CURRENT_TILE))	//=false
	//GSLogDebugInfo(GSTile.GetCargoAcceptance(CURRENT_TILE, CARGO_PAXID,1,1) 	//=0

	// Create spiral walker
	local SPIRAL_WALKER = SpiralWalker()
	// Set spiral walker on town center tile, always a road
	SPIRAL_WALKER.Start(CURRENT_TILE)

	// Create try counter
	local TRIES = 0

	// Loop till list count matches house count
	while(TOWN_HOUSE_LIST_COUNT < (TOWN_HOUSE_COUNT * HOUSE_COUNT_FACTOR) && TRIES < MAX_TRIES){
		// Inc tries
		TRIES++

		// Walk one tile
		SPIRAL_WALKER.Walk()
		// Get tile
		CURRENT_TILE = SPIRAL_WALKER.GetTile()

		// Debug sign
  		if (GSController.GetSetting("log_level") >= 4)GSSign.BuildSign(CURRENT_TILE,"" + TRIES)

		// Debug msgs
		if (GSController.GetSetting("log_level") >= 4) GSLog.Info("---" + TRIES + "---")
		if (GSController.GetSetting("log_level") >= 4) GSLogDebugInfo("In current town inf: " + GSTile.IsWithinTownInfluence(CURRENT_TILE)
		if (GSController.GetSetting("log_level") >= 4) GSLog.Info("Town authority id: " + GSTile.GetTownAuthority(CURRENT_TILE))
		if (GSController.GetSetting("log_level") >= 4) GSLog.Info("Buildable: " + GSTile.IsBuildable(CURRENT_TILE))
		if (GSController.GetSetting("log_level") >= 4) GSLog.Info("Owner id: " + GSTile.GetOwner(CURRENT_TILE))
		if (GSController.GetSetting("log_level") >= 4) GSLog.Info("Station: " + GSTile.IsStationTile(CURRENT_TILE))
		if (GSController.GetSetting("log_level") >= 4) GSLogDebugInfo("Passenger acceptance: " + GSTile.GetCargoAcceptance(CURRENT_TILE, CARGO_PAXID,1,1)

		// If not the current town, continue
		if (GSController.GetSetting("log_level") >= 4) if (GSTile.IsWithinTownInfluence(CURRENT_TILE,TOWN_ID) == false)
		// If not a ??? (non town center buildings are always 65535)
		if (GSTile.GetTownAuthority(CURRENT_TILE) != TOWN_ID) continue
		// If buildable, continue
		if (GSTile.IsBuildable(CURRENT_TILE) != false) continue
		// If owned by anyone, continue
		if (GSTile.GetOwner(CURRENT_TILE) != -1) continue
		// If station, continue
		//if (GSTile.IsStationTile(CURRENT_TILE) != false) continue
		// If industry, continue
		if (IsIndustry(CURRENT_TILE) != false) continue

		// Get passenger acceptance
		CARGO_COUNTER = GSTile.GetCargoAcceptance(CURRENT_TILE, CARGO_PAXID,1,1,0)

		// If the current tile accepts passengers
		if (GSTile.GetCargoAcceptance(CURRENT_TILE, CARGO_PAXID,1,1,0) > 0){
			// Add the tile
			TOWN_HOUSE_LIST.AddTile(CURRENT_TILE)

			// Inc counter
			TOWN_HOUSE_LIST_COUNT++

			// Debug sign
  			if (GSController.GetSetting("log_level") >= 4) GSSign.BuildSign(CURRENT_TILE,"H: " + TOWN_HOUSE_LIST_COUNT)
  			}
	}


	//GSLog.Info("Created list of " + TOWN_HOUSE_LIST.Count() + " of " + TOWN_HOUSE_COUNT + " houses in town " + GSTown.GetName(TOWN_ID))

	// Return the list
	return TOWN_HOUSE_LIST
}
*/

// Error handler function
function IndustryConstructor:: ErrorHandler(){
	// Get error
	local ERROR = GSError.GetLastError()

	// Check if error is not nothing
	if (ERROR == GSError.ERR_NONE) return

	// Error category
	 switch(GSError.GetErrorCategory()){
	case GSError.ERR_CAT_NONE:
		GSLog.Error("Error not related to any category.")
		break
	case GSError.ERR_CAT_GENERAL:
		GSLog.Error("Error related to general things.")
		break
	case GSError.ERR_CAT_VEHICLE:
		GSLog.Error("Error related to building / maintaining vehicles.")
		break
	case GSError.ERR_CAT_STATION:
		GSLog.Error("Error related to building / maintaining stations.")
		break
	case GSError.ERR_CAT_BRIDGE:
		GSLog.Error("Error related to building / removing bridges.")
		break
	case GSError.ERR_CAT_TUNNEL:
		GSLog.Error("Error related to building / removing tunnels.")
		break
	case GSError.ERR_CAT_TILE:
		GSLog.Error("Error related to raising / lowering and demolishing tiles.")
		break
	case GSError.ERR_CAT_SIGN:
		GSLog.Error("Error related to building / removing signs.")
		break
	case GSError.ERR_CAT_RAIL:
		GSLog.Error("Error related to building / maintaining rails.")
		break
	case GSError.ERR_CAT_ROAD:
		GSLog.Error("Error related to building / maintaining roads.")
		break
	case GSError.ERR_CAT_ORDER:
		GSLog.Error("Error related to managing orders.")
		break
	case GSError.ERR_CAT_MARINE:
		GSLog.Error("Error related to building / removing ships.")
		break
	case GSError.ERR_CAT_WAYPOINT:
		GSLog.Error("Error related to building / maintaining waypoints.")
		break
	 default:
      	GSLog.Error("Unhandled error category!" + GSError.GetErrorCategory())
      	break
	 }

	// Errors
	 switch(ERROR){
	 case GSSign.ERR_SIGN_TOO_MANY_SIGNS:
	 	GSLog.Error("Too many signs!")
	 	break
	 
	 default:
      	GSLog.Error("Unhandled error: " + GSError.GetLastErrorString())
      	break
	 }
}

// Function to check if tile is industry, returns true or false
function IsIndustry(TILE_ID) { return (GSIndustry.GetIndustryID(TILE_ID) != 65535) }

// Function to valuate town by dist from edge
function GetTownDistFromEdge(TOWN_ID){
	return GSMap.DistanceFromEdge(GSTown.GetLocation(TOWN_ID))
}